#!/usr/bin/python

#
#	p a s s w o r d R e p o r t . p y
#

#	Created on September 2, 2005 by Simon Brown
#

# 09/02/05	Created the shell of program using piece from reply.py
#			Created create_report & email_report routines
# 09/03/05	Defined method names for dbPassChange & implemented most of its methods
# 09/04/05	Created new dscl object based on dscl2.py
#			More work on implementing the dbPassChange class
# 09/05/06	Added the create_report & time_to_day functions
#			Shortened name of ConvertTimeToDaysBeforeExpiration to timeToDaysBeforeExpiration

import os

kMessagePath = "reportEmail.txt"
kPasswordServerLog = "/Library/Logs/PasswordService/ApplePasswordServer.Server.log.8"
kSecondsPerDay = 24 * 60 * 60
kSendmail = "/usr/sbin/sendmail"

kHourToRunPassCheck = 0			# When to check for expired passwords & email out reminders.
kHourToRunUserCheck = 18		# When to synchronize our user DB with the System's user DB.
kDaysBeforeExpiration = 30

kSecondsBeforeExpiration = kDaysBeforeExpiration * kSecondsPerDay

#
#	e m a i l _ r e p o r t
#

def email_report (inReportStr):
	"""Email out report, using headers at kMessagePath with inReportStr as body"""
	
	global kSendmail,kMessagePath
	
	# get the email message from a file
	try
		f = open(kMessagePath, 'r')
	on IOError:
		print "Error: the file " + kMessagePath + " not found"
		return False
		
	reportHeader = f.read()
	f.close()

	# open a pipe to the mail program and
	p = os.popen(kSendmail+" -t", 'w')

	# write the data to the pipe
	p.write(reportHeader)
	p.write(inReportStr)
	
	exitcode = p.close()
	return exitcode == False

#
#	c m p _ t i m e s _ i n _ l i s t
#

def cmp_times_in_list (x, y):
	return x[1]-y[1]


#
#	d b P a s s C h a n g e
#

class dbPassChange:
	# This database has a user ID as its key, and the value field is the last
	# recorded date (in ASCII text) of a password change for that user ID.

	# METHODS
	#	close (self)
	#	getChangeDate (self, inUser)
	#	getEveryUser (self)
	#	getDaysBeforeExpiration (self, inAccount)
	#	timeToDaysBeforeExpiration (self, inTime)
	#	setChangeTime (self, inAccount, inTime)
	#	stripUnknownUsers (self, inKnownUsers)
	#	time_to_day (self, inTime)

	dbPassChangePath = "/var/sibrpasschange/"
	dbRef = False

	def __init__ (self):
		self.dbRef = anydbm.open (self.dbPassChangePath, 'c')

	#
	#	c l o s e
	#
	
	def close (self):
		"""Close our database file."""
		self.dbRef.close()

	#
	#	g e t _ d a y s _ b e f o r e _ e x p i r a t i o n
	#

	def get_days_before_expiration (self, inTime)
		"""Convert time (in seconds) of last password change to no. of days until password expires."""
		
		global kSecondsPerDay,kSecondsBeforeExpiration

		# convert now to seconds @ midnight today.
		now = self.time_to_day (time.time())
		
		# Only do calculation if we didn't get a time that has already expired.
		if (inTime + kSecondsBeforeExpiration) < now:
			# This will return negative #'s when user is beyond expiration.
			return int ((kSecondsBeforeExpiration + inTime - now) / kSecondsPerDay)

		# Time value for an unknown user?
		elif inTime == -1
			return "unknown"
		else:
			return "expired"

	#
	#	get_expiring_in (inDays)
	#
	
	# Return a list containing of every user account expiring in the given number of days.
		
	def get_expiring_in (self,inDays):

		
		userList = self.get_every_user()
		expireList = []
		
		for user in userList:
			if self.get_days_before_expiration (user[1]) == inDays:
				expireList.append (user[0])
		
		return expireList

	#
	#	g e t C h a n g e D a t e
	#

	def getChangeDate (self,inUser):
	
		try:
			# Look for record with matching account ID. If successful, we'll return the
			# the date as # of seconds since last epoch.
			lastChange = self.dbRef [inUser]

		except KeyError:
			# We came here b/c the account does not exist in db, which means
			# that we have never sent seen a password change for this user.
			lastChange = 0
	
		return lastChange

	#
	#	g e t _ e v e r y _ u s e r
	#
	
	# Return a list containing every user ID with the time of last recorded
	# time of password change.
	
	def get_every_user(self):
		userAndDateList = []
		
		# Add the two tuple data from DB as a nested list.
		userAndDateList.append (dbRef.first())
		
		# Add all the remaining items in list, one by one.
		for i in xrange(1, len(dbRef)):
			userAndDateList.append (dbRef.next())
		
		return userAndDateList


	#
	#	s e t _ p w _ c h a n g e s
	#
	
	# Updates the database from the given list containing account ID's and their password change dates
	
	def set_pw_changes(self,inChangeList):
		
		# Add all the remaining items in list, one by one.
		for user in inChangeList:
			self.set_user_pw_change (user[0],user[1])
		

	#
	#	g e t D a y s B e f o r e E x p i r a t i o n
	#

	def getDaysBeforeExpiration (self, inAccount):
		"""Return # of days before a user's password expires, 'unknown', or 'expired'."""
		
		global kSecondsBeforeExpiration, kSecondsPerDay
		
		try:
			# Look for record with matching account & sender address.
			# If it's successful, it will have the time (# of seconds since last epoch)
			# that the last reply was sent.

			lastChange = self.getChangeDate (inAccount)
	
		except KeyError:
			# We came here b/c the account's key does not exist,
			# which means that we have never sent seen a password change for this user.
			lastChange = -1
	
		daysBeforeExpiration = self.timeToDaysBeforeExpiration (lastChange)

		print "User="+inAccount, "lastChange="+lastChange,"daysBeforeExpiration="+daysBeforeExpiration

		return daysBeforeExpiration


	#
	#	p a s s w o r d _ r e p o r t
	#

	def password_report (self):
		"""Returns two lists, first users sorted by ID, second users sorted by expiration time"""
		# Get the raw data first.
		userList = self.get_every_user()
		listByUser = []
		listByExpiration = []
		
		for user in userList:
			# Add field containing calculated time before user's password expires.
			user.append (self.convertTimeToDaysBeforeExpiration(user[1]))
			listByUser.append (user)
		
		listByUser.sort()

		listByExpiration = listByUser
		listByExpiration.sort (cmp_times_in_list)
	
		return listByUser, listByExpiration


	#
	#	s t r i p U n k n o w n U s e r s
	#

	def stripUnknownUsers (self,inKnownUsers):
		"""Removes users not in given list from DB."""
	
		# First, get every user that is currently in our DB.
		dbUsers = self.dbRef.keys()
		
		for user in dbUsers:
			try:
				# We don't really care about the result here -- we just want to know if user is present or not.
				inKnownUsers.index (user)
			except ValueError:
				# User not present in system user list, so we'll remove it from ours too.
				del self.dbRef.del (user)
				print "User "+user+" deleted from DB"
			

	#
	#	s e t _ u s e r _ p w _ c h a n g e _ d a t e
	#

	def set_user_pw_change_date (self,inAccount, inTime):
	
		# Create a key that is based on the user's account plus the sender's
		# email address. Then store the time that the reply was sent.
		self.dbRef [self.getDBKey (inAccount, inSenderAddress)] = str (int (time.time()))


	#
	#	t i m e _ t o _ d a y
	#
	
	# Round down the given time (in seconds since epoch) to its day
	
	def time_to_day (self, inTime):
	
		t = time.localtime()
		tRounded = (t[0], t[1], t[2], 0, 0, 0, 0, 0, t[8])
		return time.mktime (tRounded)
		

#
# end class dscl	
#

class dscl:

	def list_users ():
		# Must be run as root: if not, a user & password will be required.
		usersStr = commands.getoutput ("dscl localhost -list /LDAPv3/127.0.0.1/Users")
		return usersStr.splitlines()
	list_users = staticmethod (list_users)
	
#
# end class dscl
#


class PasswordServerLog:

	kPathToLog = "/Library/Logs/PasswordService/ApplePasswordServer.Server.log"
	
	def get_change_events ():
		eventsStr = commands.getoutput ("awk -F' '  '/CHANGE/ {print $1,$2,$3,substr ($13,1,length($13)-1)}' " + self.kPathToLog)
		return eventsStr.splitlines()
		
	get_change_events = staticmethod (get_change_events)
	

#
# end class PasswordServerLog
#


def main()
	"""Main program loop."""

	hour = time.strftime('%H')
	
	# Time to check for any passwords close to expiring?
	if hour == kHourToRunPassCheck:
		reportStr = create_report()
		email_report (reportStr)
		
	if hour = kHourToRunUserCheck:
		# Get a list of every user in system.
		systemUsers = dscl.list_users()
		
		# Remove any users from our DB that aren't in the System's DB
		userDB.stripUnknownUsers (systemUsers)
	
	
	