#!/usr/bin/python

#
#	p w M o n i t o r . p y
#

#	Created on September 2, 2005 by Simon Brown
#

# 09/02/05	Created the shell of program using piece from reply.py
#			Created create_report & email_report routines
# 09/03/05	Defined method names for dbPassChange & implemented most of its methods
# 09/04/05	dscl: created new object based on dscl2.py
#			dbPassChange: more work on implementing class
# 09/06/05	Added the create_report & time_to_day functions
#			Shortened name of ConvertTimeToDaysBeforeExpiration to timeToDaysBeforeExpiration
# 09/07/05	b/c of Daylight Savings Time decided to not store times in Python's seconds since epoch.
#			Stripped down a version of passwordReport.py to mostly just record dates of 
#			First working version!
# 09/08/05	Cleaned up the debug output.
#			Added & tested the strip_unknown_users method.
# 09/09/05	Figured out a way to go back to using "seconds since epoch" to store
#			dates. This makes a number of calculations easier then working with
#			date strings.
#			Added the days_before_expiration, time_to_day, email_text, and password_report functions.
# 09/10/05	Changing name to pwMonitor
#			password_report: changed report formatting
#			Added configuration values needed to send email reminders.
#			Updated comments.
#			synchronize_users: added method
#			cmp_times_in_list: modified to work with unknown dates.
# 09/11/05	password_report_list: fixed prob. w/  making reference instead of copy of list,
#			and put list sorted by days before list sorted by user.
#			pw_expiring_in: added new dbPassChange method.
#			email_text: changed to allow input of headers and body as str input.
#			print_user_item: added to password_report to remove redundant code.
#			email_text: changed again, made it accept one or more string or path items.
#			password_report: changed string to represent unknown dates to "--/--/--".
#			Email reminders now working.
# 09/13/05	Updated comments
#			return_str_or_file: Now only add \n to end when none is present.
#			print_user_item: removed extraneous assignment.
#			synchronize_users: now check for input of empty user list.
#			Changed name of database file from sibr.pwchange.db to pwMonitor.db
#			get_change_events: replaced the awk based code w/ one using python's re lib.
# 09/14/05	get_change_events: Changed from triple-qouted to r" text for regex pattern.
# 09/27/05	kRemSubject: edited to add "Urgent" to the subject
#			email_reminders: now returns string on who was emailed a reminder.
#			main: now email report after sending out email reminders
#			main: now add output from email_reminders to reportStr
# 10/01/05	password_report: fixed typo in variable name
#			Beta test begins.
# 10/02/05	time_to_day: fixed bug where current time was being used instead of inTime.
# 10/03/05	get_change_events: not catching some events at end of log when log file was
#			rolled (duh!). Back to using a modified version of awk based solution that
#			checks current + two of the rolled password server logs. The upside on this 
#			change is that we only need to check logs once a day now vs. once an hour.
#			main: now call synchronize_users before calling password_report_lists.
#			get_every_user: replaced for loop w/ a list comprehension for speed.
# 10/05/05	password_report: replaced for loops w/ list comprehensions
#			password_report: using lists to create report string instead of str concatenation.
# 10/08/05	Made some minor changes to program output.
#			main: removed now unneeded calculation of current hour.
# 10/17/05	set_pw_changes: now returns a report string.
#			main: changed to accomodate report string from set_pw_changes
#			password_report: changed header on report text
#			Changed dscl calls to use the kDsclPath so that we now specify the full cmd. path.
# 01/04/06	get_change_events: now sort by account name & remove dup events

#	NOTES:
#		Any advantages to using the datetime type instead of time?

import commands,os,string,sys,time
import anydbm

kReportMsgHeaderPath = "/Users/Shared/sibr.pwMonitor/reportMsgHeader.txt"

# We watch this log looking for password change events.
# The glob pattern matches the current log and the two most recently rolled logs.
kPasswordServerLogs = "/Library/Logs/PasswordService/ApplePasswordServer.Server.log{,.1,.2}"

kDsclPath = "/usr/bin/dscl"
kSendmailPath = "/usr/sbin/sendmail"
kSecondsPerDay = 24*60*60
kUnknownDate = -1				# Value used when date of last password change is unknown.

#	Headers & text for reminder and report messages

kRemSubject = 'Subject: Urgent! Your email password will be expiring in '
kRemFrom = 'From: Password Monitor <admins@domain.com>'
kRemPrecedence = 'Precedence: bulk'
kRemBody = '/Users/Shared/sibr.pwMonitor/reminderMsgBody.txt'

#	Domain name

kDomain = 'domain.com'

#	Task Switches

kSendReminders = True			# Email reminder message to users that their password will be expiring?
kSendReports = True				# Email password aging reports?

#	Timing of password expiration & reminders

kDaysBeforeExpiration = 30		# How long between forced password resets?
kFirstWarningDays = 4			# No. days before first email warning of password expiration.
kSecondWarningDays = 2			# 2nd warning of password expiration.

# Subtracting 1 hour from this calculation to allow for "Fall Back" at end of
# Daylight Savings Time. This will give us a one hour slop when doing comparisons,
# which works b/c we are rounding all times to seconds at midnight.
kSecondsBeforeExpiration = kDaysBeforeExpiration * kSecondsPerDay - 60*60


#
#	c m p _ d a y s _ i n _ l i s t
#

def cmp_days_in_list (listX, listY):
	"""Compare the third item (time in seconds) of given parameters, treating -1
	(unknown date) as the highest value."""
	
	# Short circuit the result if dealing with unknown date values
	if listX[2] == -1: return 1
	if listY[2] == -1: return -1

	# Compare the two time items that are 2nd item in list.
	return listX[2] - listY[2]


#
#	e m a i l _ t e x t
#

def email_text (*inStrOrPath):
	"""Sends an email using sendmail. The input is one or more parameters, which may either
	be a string or a file path. File path parameters must begin with the "/" character. If
	either the given string or the text from the file is missing a terminating \n then it
	will be added. May throw an IOError if there was a problem reading in the file."""
	
	def return_str_or_file (strOrPathItem):
	
		# If first char is forward slash then we'll treat the string as a file path.
		if strOrPathItem[0] == '/':
			# Attempt to read in the message header in the file we were given the path to.
			# open file w/ read-only access.
			f = open (strOrPathItem,'r')
			
			# Read in the text and close the file.
			strOrPathItem = f.read()
			f.close()
		
		# Make sure we terminate each line.
		if strOrPathItem [-1] != '\n':
			return strOrPathItem + "\n"
		else:
			return strOrPathItem

			
	# Open a connection to the sendmail command in the shell.
	#  -i  don't treat a . on a line as end of text
	#  -t  extract recipients from text of message.
	
	p = os.popen(kSendmailPath + ' -i -t', 'w')
	
	for strItem in inStrOrPath:
		# Read in file if necessary, then output string to the sendmail process.
		# If it was a file & there was an error reading it in we'll catch that here.
		try:
			# return_str_or_file may throw an exception.
			x = p.write (return_str_or_file (strItem))

		except IOError:
			print 'Error: could not open the file', strItem
			return False

	# Close our connection to the sendmail/shell process.
	exitCode = p.close()
	
	return exitCode == False

#
#	e m a i l _ r e m i n d e r s
#

def email_reminders (inExpiringUsers, inDaysBeforeExpire):
	"""Emails out reminders messages to users. We are given a list of users (which may be empty)
	and how long before their password expires. A string is returned with a list of every
	user that was sent a reminder."""
	
	logStr = 'email_reminders: ' + str (inDaysBeforeExpire) + ' days\n'

	for user in inExpiringUsers:
		# Piece together a message header and its body.
		email_text ( kRemSubject + str (inDaysBeforeExpire) + ' days', \
					 kRemFrom, \
					 'To: ' + user + '@' + kDomain, \
					 kRemPrecedence + '\n\n', \
					 kRemBody )

		logStr += '\t' + user + '\n'

	print logStr
	return logStr

#
#	d b P a s s C h a n g e
#

class dbPassChange:
	# This database has a user ID as its key, and the value field is the last
	# recorded date (in seconds) of a password change for that user ID.

	# This will be the file containing the user & date information database managed by this
	# object. A new empty database file will be created if none is found.
	dbPassChangePath = "/var/db/pwMonitor"
	
	# This will hold the reference to the database after an object has been initialized.
	dbRef = False

	def __init__ (self):
		# Open the database file, creating a new one if none found.
		self.dbRef = anydbm.open (self.dbPassChangePath, 'c')

	#
	#	c l o s e
	#
	
	def close (self):
		"""Close our database file."""
		self.dbRef.close()
	
	#
	#	d a y s _ b e f o r e _ e x p i r a t i o n
	#

	# Calculates the number of days before the given time (rounded to the day @ midnight)
	# will reach its expiration. For times that have already expired the string "expired"
	# is returned. If given -1 (used if a record is added but the last password change is
	# not known) then "unknown" is returned.
	
	def days_before_expiration (self, inTime):
		"""When given the seconds since midnight of last password change, gives # days b/f pw expires."""
		
		# Unknown dates are returned unchanged.
		if inTime == kUnknownDate:
			return kUnknownDate
			
		# Convert time to seconds @ midnight today.
		today = self.time_to_day (time.time())

		# Only do calculation if we didn't get a time that has already expired.
		if (inTime + kSecondsBeforeExpiration) > today:
			# This will return negative #'s if given date that has already expired.
			return int ((kSecondsBeforeExpiration + inTime - today) / kSecondsPerDay)
		else:
			# We'll make the assumption (poss. incorrect) that this is an expired account.
			return 0

	#
	#	g e t _ e v e r y _ u s e r
	#
		
	def get_every_user(self):
		"""The database is queried for every user. A list containing two items for each
		user is returned: the user ID, and the date of the last recorded password change."""

		# Get a list of every user ID in database.
		usersList = self.dbRef.keys()
		userDB = self.dbRef
		
		# Add the dates for every user account and add both items to a new list.
		#for user in usersList:
			# The key's values is always stored as a string, so we use int to
			# convert the key value back to an integer value.
			#userAndDateList.append ([user,int (self.dbRef[user])])
			
		userAndDateList = [ [user,int (userDB[user])] for user in usersList ]
		
		return userAndDateList


	#
	#	p a s s w o r d _ r e p o r t
	#

	def password_report (self, userList, dateList):
		"""Returns a string with a formatted & sorted report, listing each user alphabetically
		in one section, and by days before expiration in another."""

		def print_user_item (user):
		
			# User name, padded on the right in a 20 character field.
			userStr = string.ljust (user[0],20)
			
			# The date of last password change, in the format mm/dd/yy.
			if user[1] > 0:
				userStr += string.ljust (time.strftime ('%m/%d/%y', time.localtime (user[1])),14)
			else:
				userStr += string.ljust ('--/--/--',14)
				
			# No. of days before user's password expires.
			if user[2] > 0:
				userStr += str (user[2])
			elif user[2] == 0:
				userStr += 'expired'
			else:
				# Should only get here if -1.
				userStr += 'unknown'
				
			return userStr

		# Note: now use lists to store each line of strings for speed, although
		# it does add an extra step at end to join together the items into string.

		# Header for first section of report.
		reportLines = ["Password Aging: sorted by date\n"]
		
		# Append the data for each user to the string.		
		reportLines += [ print_user_item (user) for user in dateList ]

		reportLines += ["\nPassword Aging: sorted by user\n"]

		# Append data for each user, sorted by user
		reportLines += [ print_user_item (user) for user in userList ]

		# Join together the lines into a giant string, using \n as the seperator
		# for each item in the list.
		return '\n'.join (reportLines) + '\n'


	#
	#	p a s s w o r d _ r e p o r t _ l i s t s
	#

	def password_report_lists (self):

		# Get every user in the DB with the date of their last password change.
		userList = self.get_every_user()
		
		listByUser = []
		
		# Create new list by iterating through the one provided by get_every_user.
		for user in userList:
			# Add an additional field with calculated time before user's password expires.
			user.append (self.days_before_expiration (int (user[1])))
			listByUser.append (user)
		
		# Sort list by name.
		listByUser.sort()
				
		# We'll use the first list we generated as the starting point for the 2nd list.
		listByExpiration = listByUser[:]
		
		# Only add user to list if password has expired.
		# listByExpiration = [ user for user in listByUser if user[2] == 0 ]
		
		# Sort the list, using the 2nd item in list as its criteria.
		listByExpiration.sort (cmp_days_in_list)
		
		return listByUser, listByExpiration

	#
	#	s e t _ p w _ c h a n g e s
	#
	
	# Updates the database from the given list containing account ID's and their password change dates
	
	def set_pw_changes(self,inChangeList):
		
		reportStr = "Changed Passwords:\n\n"

		# Add all the remaining items in list, one by one.
		for user in inChangeList:
			# Convert the date string into time tuple, and then to time in seconds.
			timeInSeconds = time.mktime (time.strptime (user [0],"%b %d %Y"))

			# Save the date as time in seconds in database.
			self.dbRef [user[1]] = str (int (timeInSeconds))

			reportStr += '%-20s %s\n' % (user[1], user [0])

		return reportStr

	#
	#	s y n c h r o n i z e _ u s e r s
	#

	def synchronize_users (self,inKnownUsers):
		"""Removes users not in given list from DB."""
		
		# If we got an empty list somehow we could wipe the whole database!
		if inKnownUsers == []:
			print "User list is empty! Database not updated."
			print "Bad call to dscl?"
			return

		# First, get every user that is currently in DB.
		dbUsers = self.dbRef.keys()
		
		print "Checking for users not in directory"
		for user in dbUsers:
			try:
				# We don't really care about the result here -- we just want to know if user is present or not.
				inKnownUsers.index (user)
				#print "OK:",user
			except ValueError:
				# User not present in system user list, so we'll remove it from ours too.
				print "DELETE:", user
				del self.dbRef [user]

		print
		print "Checking for users in directory but not database"
		
		for user in inKnownUsers:
			try:
				# We don't really care about the result here -- we just want to know if user is present or not.
				dbUsers.index (user)
				#print "OK:",user
			except ValueError:
				# User present in system user list, so we'll add it to ours too.
				print "add:", user
				
				# Give the new user a value indicating the last password change time is unknown.
				self.dbRef [user] = str(kUnknownDate)

	#
	#	t i m e _ t o _ d a y
	#

	def time_to_day (self,inTime):
		"""Round inTime down to the seconds at the beginning of the inTime's day. Does NOT adjust for DST."""
		
		# Split the time into seperate date & time items.
		t = time.localtime(inTime)

		# Only use the date & DST mode items.
		# NOTE: could the DST item be used to fix Daylight Savings Time issues???
		
		tRounded = (t[0], t[1], t[2], 0, 0, 0, 0, 0, t[8])
		
		# Join all the date & time items back up into time since epoch.
		return time.mktime (tRounded)

	#
	#	p w _ e x p i r i n g _ i n
	#
	
	# Takes a report list sorted by pw expiration containing user name, pw change time, 
	# and days before expiration. Returns users whose pw is good for the # of days
	# given in inDaysExpiringIn.
	
	def pw_expiring_in (self, inReportList, inDaysExpiringIn):

		expiringUsers = []
		
		print "Looking for pw expiring in",inDaysExpiringIn,'days'
		
		# Loop through user list looking for users expiring in the given time.
		# Could use map command here?
		for user in inReportList:
			if user[2] == inDaysExpiringIn:
				expiringUsers.append (user[0])

		return expiringUsers

#
# end class dbPassChange
#

#
#	d s c l	
#

class dscl:

	#
	#	l i s t _ u s e r s
	#

	def list_users ():
		"""Returns a list all users known by Open Directory."""
		
		# Must be run as root: if not, a user & password will be required.
		dsclResult,usersStr = commands.getstatusoutput (kDsclPath+" localhost -list /LDAPv3/127.0.0.1/Users")
		
		# Return an empty list if dscl has an error.
		if dsclResult != 0: return []
		
		# Convert the string of line delimited data to a list.
		return usersStr.splitlines()
		
	list_users = staticmethod (list_users)
	
#
# end class dscl
#


#
#	P a s s w o r d S e r v e r L o g
#

class PasswordServerLog:

# An example of a change password event:
# Sep  7 2005 08:34:57    CHANGEPASS: {0x427924994ae3b6500000001200001012, lking} changed password for user {0x427924994ae3b6500000001200001012, lking}
	
	def get_change_events ():
		"""Extract password change events from the Password Server log. Return a list
		containing each account name and the date (as string) when they changed their
		password."""
		
		# Open up the Password Server log file. We won't catch any IOErrors here,
		# since not being able to open the log would be a serious error.
		#f = open (kPasswordServerLog,'r')
	
		# Read in the text and close the file.
		#eventsStr = f.read()
		#f.close()

		# FOR DEBUGGING
		#eventsStr += 'Sep  2 2005 08:34:57    CHANGEPASS: {0x427924994ae3b6500000001200001012, sbrown} changed password for user {0x427924994ae3b6500000001200001012, sbrown}\n'
		#eventsStr += 'Sep  2 2005 08:34:57    CHANGEPASS: {0x427924994ae3b6500000001200001012, jriker} changed password for user {0x427924994ae3b6500000001200001012, jriker}\n'
		
		# Grab everything from beginning of line up to the year: this should be one item.
		# Next, get one or more alphanumerics (should be user ID), terminated by brace and EOL.
		# $ not working for EOL, having to use \n instead.
		#return re.findall (r"(.+ \d\d\d\d) .+, (\w+)\}\n",eventsStr)
	
		eventsStr = commands.getoutput ("""/usr/bin/awk -F' '  '/changed passw/ {print $1,$2,$3"\t"substr ($13,1,length($13)-1)}' """ + kPasswordServerLogs + " | sort -u")
		eventLines = eventsStr.splitlines()

		# Seperate the date & user ID for every line of output from awk script.
		eventItems = [ line.split("\t") for line in eventLines ]

		return eventItems

	get_change_events = staticmethod (get_change_events)
	
#
# end class PasswordServerLog
#

#
#	m a i n
#

def main():
	"""Main program loop."""

	print "Start:",time.asctime()
	print
	
	# Get a list of every password change event currently in Password Server log.
	eventItemsList = PasswordServerLog.get_change_events()

	# Create the user password database object & open the database file.
	userDB = dbPassChange()
	
	# Remove deleted accounts & add new accounts.
	userDB.synchronize_users (dscl.list_users())

	print

	# Update dates for users that have changed their password.
	reportStr = userDB.set_pw_changes (eventItemsList) + '\n'

	# Get the raw data needed for the password report & email reminders.
	userList, dateList = userDB.password_report_lists()
	
	# Create the formatted report.
	reportStr += userDB.password_report(userList, dateList)
	
	print
	print reportStr

	# SEND REMINDERS?
	if kSendReminders:
		reportStr += '\n' + \
					 email_reminders (userDB.pw_expiring_in (dateList,kFirstWarningDays),kFirstWarningDays) \
					 + '\n'
		reportStr += email_reminders (userDB.pw_expiring_in (dateList,kSecondWarningDays),kSecondWarningDays) \
					 + '\n'

	# SEND REPORTS?
	if kSendReports:
		email_text (kReportMsgHeaderPath,reportStr)	
	
	# We're done with the database.
	userDB.close()
	
	print
	print "Done:",time.asctime(),"\n"

#
#
#

main()

#
#
#
	
	
